/***************************************************************************

    filename             : TRfn.C
    description          : 
    created              : Tue Jul 15 2003
    version              : 1.0
    author               : Micha D. Niskin
    email                : mniskin@yahoo.com

    Copyright (C) 2003 Micha D. Niskin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

 ***************************************************************************/


#include "TRfn.h"


ClassImp(TRfn)

//-------------------------------------------------------------------------

TRfn::TRfn()
{
   fChild1  = fChild2 = 0;
   fOp      = kNONE;
   fA       = 0;
   fB       = 0;
   fC       = 0;
   fDR_DTh  = 0;
   fDR_DPh  = 0;
   fIsValid = kTRUE;
   fIsNegated = kFALSE;
}

//-------------------------------------------------------------------------

TRfn::TRfn(Double_t a, Double_t b, Double_t c)
{
   fChild1  = fChild2 = 0;
   fOp      = kPRIMITIVE;
   fA       = a;
   fB       = b;
   fC       = c;
   fDR_DTh  = a / TMath::Sqrt(a*a + b*b);
   fDR_DPh  = b / TMath::Sqrt(a*a + b*b);
   fIsValid = kTRUE;
   fIsNegated = kFALSE;
}

//-------------------------------------------------------------------------

TRfn::TRfn(const TRfn& rfn)
{
   if ((rfn.GetOp() == kAND) || (rfn.GetOp() == kOR)) {
      fChild1 = new TRfn((const TRfn&) *(rfn.GetChild1()));
      fChild2 = new TRfn((const TRfn&) *(rfn.GetChild2()));
   } else {
      fChild1 = fChild2 = 0;
   }
   fOp      = rfn.GetOp();
   fA       = rfn.GetA();
   fB       = rfn.GetB();
   fC       = rfn.GetC();
   fDR_DTh  = rfn.GetDR_DTh();
   fDR_DPh  = rfn.GetDR_DPh();
   fIsValid = rfn.IsValid();
   fIsNegated = rfn.IsNegated();
}

//-------------------------------------------------------------------------

TRfn::~TRfn()
{
   if (fChild1)
      delete fChild1;
   if (fChild2)
      delete fChild2;
}

//-------------------------------------------------------------------------

TRfn TRfn::operator= (TRfn rfn)
{
   if ((rfn.GetOp() == kAND) || (rfn.GetOp() == kOR)) {
      fChild1 = new TRfn((const TRfn&) *(rfn.GetChild1()));
      fChild2 = new TRfn((const TRfn&) *(rfn.GetChild2()));
   } else {
      fChild1 = fChild2 = 0;
   }
   fOp      = rfn.GetOp();
   fA       = rfn.GetA();
   fB       = rfn.GetB();
   fC       = rfn.GetC();
   fDR_DTh  = rfn.GetDR_DTh();
   fDR_DPh  = rfn.GetDR_DPh();
   fIsValid = rfn.IsValid();
   fIsNegated = rfn.IsNegated();

   return *this;
}

//-------------------------------------------------------------------------

TRfn TRfn::operator! ()
{
   TRfn r((const TRfn&) *this);
   r.SetIsNegated(!r.IsNegated());
   return r;
}

//-------------------------------------------------------------------------

TRfn TRfn::operator& (TRfn a)
{
   TRfn r = *this;
   r &= a;
   return r;
}

//-------------------------------------------------------------------------

TRfn TRfn::operator| (TRfn a)
{
   TRfn r = *this;
   r |= a;
   return r;
}

//-------------------------------------------------------------------------

void TRfn::operator&= (TRfn a)
{
   if (!fIsValid || !a.IsValid()) {
      fIsValid = kFALSE;
      return;
   } else if (fOp == kNONE) {
      (*this) = a;
      return;
   }

   TRfn *copy1 = new TRfn((const TRfn&) *this);
   TRfn *copy2 = new TRfn((const TRfn&) a);

   if (fChild1) 
      delete fChild1;
   if (fChild2) 
      delete fChild2;

   fChild1 = copy1;
   fChild2 = copy2;
   fOp     = kAND;
}

//-------------------------------------------------------------------------

void TRfn::operator|= (TRfn a)
{
   if (!fIsValid && !a.IsValid()) {
      return;
   } else if (fOp == kNONE) {
      (*this) = a;
      return;
   }

   TRfn *copy1 = new TRfn((const TRfn&) *this);
   TRfn *copy2 = new TRfn((const TRfn&) a);

   if (fChild1) 
      delete fChild1;
   if (fChild2) 
      delete fChild2;

   fChild1 = copy1;
   fChild2 = copy2;
   fOp     = kOR;
}

//-------------------------------------------------------------------------

Double_t TRfn::Primitive(Double_t th, Double_t ph)
{
   return (fA*th + fB*ph + fC)/TMath::Sqrt(fA*fA + fB*fB);
}

//-------------------------------------------------------------------------

Double_t TRfn::And(Double_t a, Double_t b)
{
   return (a < b) ? a : b;
}

//-------------------------------------------------------------------------

Double_t TRfn::Or(Double_t a, Double_t b)
{
   return (a > b) ? a : b;
}

//-------------------------------------------------------------------------

Double_t TRfn::operator() (Double_t th, Double_t ph)
{
   Double_t retval = kBAD_RFN;

   if (!IsValid())
      return(retval);

   if (fOp == kPRIMITIVE) {
      //fDR_DTh = DR_DTh(th, ph);
      //fDR_DPh = DR_DPh(th, ph);
      retval = Primitive(th, ph);
   } else {
      Double_t r1 = ((TRfn) *fChild1)(th, ph);
      Double_t r2 = ((TRfn) *fChild2)(th, ph);
      
      if (fOp == kAND)
         retval = And(r1, r2);
      else if (fOp == kOR)
         retval = Or(r1, r2);

      //fDR_DTh = (retval == r1) ? fChild1->GetDR_DTh() : fChild2->GetDR_DTh();
      //fDR_DPh = (retval == r1) ? fChild1->GetDR_DPh() : fChild2->GetDR_DPh();
   }

   if (IsNegated())
      retval = -retval;
   
   return(retval);
}
