/***************************************************************************

    filename             : THaRfn.C
    description          : 
    created              : Sun Jul 20 2003
    version              : 1.0
    author               : Micha D. Niskin
    email                : mniskin@yahoo.com

    Copyright (C) 2003 Micha D. Niskin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

 ***************************************************************************/


#include <fstream>
#include <snprintf.h>
#include <TString.h>
#include "THaRfn.h"


using std::ifstream;
using std::ios;
using std::string;


ClassImp(THaRfn)


//-------------------------------------------------------------------------

THaRfn::THaRfn() : TNamed(), fDate(), fGrid()
{
   fCollimator = kOUT;
   fArm        = kLEFT;
   fDy         = 0;
   fDd         = 0;
   fDR_DY      = 0;
   fDR_DD      = 0;
   fDR_DTh     = 0;
   fDR_DPh     = 0;
   fIsValid    = kFALSE;
   fGrid.SetOwner();
}

//-------------------------------------------------------------------------

   /*
    * Constructor: collimator is kIN or kOUT, arm is either kLEFT or kRIGHT,
    * and dy, dd are the grid spacings.
    */

THaRfn::THaRfn(const char* name, const char* title, ECollimator_t collimator,
               EArm_t arm, Double_t dy, Double_t dd) : TNamed(name, title), 
               fDate(), fGrid()
{
   if (!dy || !dd) {
      Error("THaRfn","Attempt to create THaRfn with zero grid spacing.");
      return;
   }
   
   fCollimator = collimator;
   fArm        = arm;
   fDy         = dy;
   fDd         = dd;
   fDR_DY      = 0;
   fDR_DD      = 0;
   fDR_DTh     = 0;
   fDR_DPh     = 0;
   fIsValid    = kTRUE;
   fGrid.SetOwner();
}

//-------------------------------------------------------------------------

   /*
    * Add the line a*theta + b*phi + c = 0 to the TRfn associated with the
    * grid point (y,dp).
    */

void THaRfn::AddLine(Double_t y, Double_t dp, Double_t a, Double_t b,
                     Double_t c)
{
   if (!fIsValid)
      return;

   TRfn r(a, b, c);
   TRBNode n(y, dp, fDy, fDd, 0);

   if (TRBNode* found = (TRBNode*) fGrid.FindObject(&n)) {
      TRfn* rfn = (TRfn*) found->fData;
      (*rfn) &= r;
   } else {
      TRfn* newrfn = new TRfn((const TRfn&) r);
      TRBNode* newnode = new TRBNode(y, dp, fDy, fDd, newrfn);
      newnode->SetOwner();
      fGrid.Add(newnode);
   }
}

//-------------------------------------------------------------------------

   /*
    * Disable the TRfn object corresponding to the grid point (y, dp).
    * This causes the TRfn to always return TRfn::kBAD_RFN (-1000).  
    */

void THaRfn::DiscardCut(Double_t y, Double_t dp)
{
   if (!fIsValid)
      return;

   TRBNode n(y, dp, fDy, fDd, 0);
   if (TRBNode* found = (TRBNode*) fGrid.FindObject(&n))
      ((TRfn*) found->fData)->SetIsValid(kFALSE);
}

//-------------------------------------------------------------------------

   /*
    * Evaluate the R-function at the given point. Negative results indicate
    * events outside of the acceptance region, positive results indicate
    * good events. The special value TRfn::kBAD_RFN (-1000) indicates an
    * error (such as the y and dp requested are outside of the grid points,
    * or the TRfn object corresponding to those grid points is invalid).
    */  

Double_t THaRfn::Eval(Double_t y, Double_t dp, Double_t th, Double_t ph)
{
   if (!fIsValid)
      return TRfn::kBAD_RFN;

   TRBNode n(y, dp, fDy, fDd, 0);
   TRBNode *found=0;
   Double_t retval=TRfn::kBAD_RFN;

   if ( (found = (TRBNode*) fGrid.FindObject(&n)) )
      retval = (((TRfn*) found->fData)->Eval(th, ph));
   else
      return(retval);

/*    [COMPUTE ERRORS THIS SECTION]
   TRBNode *next_grid_pt=0;

   n.fKey1 = found->fKey1 + fDy;
   next_grid_pt = (TRBNode*) fGrid.FindObject(&n);
   if (!next_grid_pt) {
      n.fKey1 = found->fKey1 - fDy;
      next_grid_pt = (TRBNode*) fGrid.FindObject(&n);
   }

   if (next_grid_pt)
      fDR_DY = (retval-((TRfn*)next_grid_pt->fData)->Eval(th, ph))/fDy;

   n.fKey1 = y;
   n.fKey2 = found->fKey2 + fDd;
   next_grid_pt = (TRBNode*) fGrid.FindObject(&n);
   if (!next_grid_pt) {
      n.fKey2 = found->fKey2 - fDd;
      next_grid_pt = (TRBNode*) fGrid.FindObject(&n);
   }

   if (next_grid_pt)
      fDR_DD = (retval-((TRfn*)next_grid_pt->fData)->Eval(th, ph))/fDd;

   fDR_DTh = ((TRfn*) found->fData)->GetDR_DTh();
   fDR_DPh = ((TRfn*) found->fData)->GetDR_DPh();
*/

   return(retval);
}

//-------------------------------------------------------------------------

  /*
   * Reads cut definitions from John LeRose's dat files.
   * Assumes the existence of the usual files in the directory
   * specified in the argument:
   *     1.  information.dat
   *     2.  dy-dd.dat
   *     3.  line1.dat
   *     ...
   *     14. line12.dat
   *     15. min-max.dat 
   */

void THaRfn::LoadDefs(const char* dir)
{
   fIsValid = kTRUE;

   Double_t ndum, ydum, ddum, tmin, tmax, phimin, phimax, a, b, c, rsquared;
   char filename[64], line1[256];
   TString string1;
   ifstream file;

   //----------------------------------------------------------------------

   // open the information.dat file and figure out whether the collimator
   // is IN or OUT, what the date was, and whether it's the LEFT or RIGHT arm

   // open the info file
   snprintf(filename, sizeof(filename), "%s/information.dat", dir);
   file.open((const char*) filename, ios::in);
   if (!file.is_open()) {
      Error("Dat2Root","Can't open %s: aborting",filename);
      return;
   }

   // first line is the date, in the form 'Fri Apr 26 16:30:47 2002'
   // need to convert that to '20020426163047'
   char dayname[64], month[64], time[64];
   int  day, year;
   file >> dayname >> month >> day >> time >> year;
   file.ignore(256,'\n');     // seek past the first line
   TDatime date = ParseDate((const char*) month, day, (const char*) time, year);
   SetDate((const TDatime&) date);
   
   // second line indicates whether collimator is in or out
   // need to convert to THaRfn::ECollimator_t::(kIN|kOUT)
   file.get(line1,256,'\n');  // read second line
   file.ignore(256,'\n');     // advance to third line
   string1=line1;
   string1.ToUpper();
   fCollimator = (string1.Contains(" IN")) ? kIN : kOUT;

   // third (last) line indicates whether rfn is for left or right arm
   // need to convert to THaRfn::EArm_t::(kLEFT|kRIGHT)
   file.get(line1,256,'\n'); // read last line
   string1 = line1;
   string1.ToUpper();
   fArm = (string1.Contains("LEFT")) ? kLEFT : kRIGHT;
   
   file.close();
   file.clear();  // reset errors and seek pointers

   //----------------------------------------------------------------------

   // open the dy-dd.dat file and get the two numbers which indicate the
   // distance in y and dp between adjacent grid points (grid spacing is
   // uniform)

   // open dy-dd.dat file
   snprintf(filename, sizeof(filename), "%s/dy-dd.dat", dir);
   file.open((const char*) filename, ios::in);
   if (!file.is_open()) {
      Error("Dat2Root","Can't open %s: aborting",filename);
      return;
   }

   file >> fDy >> fDd;  // set grid spacing
   file.close();
   file.clear();

   //----------------------------------------------------------------------

   // Loop through the lineX.dat files adding lines to the r-function.
   // Each line is of the form 'n y dp a b c r2', where n = 1 in file
   // line1.dat, n = 2 in line2.dat, etc., and r2 is not currently used.
   for (int j=1; j <= 12; j++) {
      snprintf(filename, sizeof(filename), "%s/line%d.dat", dir, j);
      file.open((const char*) filename, ios::in);
      if (!file.is_open()) {
      Error("Dat2Root","Can't open %s: aborting",filename);
         return;
      }

      while (file.good()) {
         file >> ndum >> ydum >> ddum >> a >> b >> c >> rsquared;
         //printf("%12f%12f%12f%12f%12f%12f\n",ndum, ydum,ddum,a,b,c);
         if (b != 0)
            AddLine(ydum, ddum, a, b, c);
      }
      file.close();
      file.clear();
   }

   //----------------------------------------------------------------------

   // Get bounding box definitions from min-max.dat file and add
   // them as cuts to r-function. A cut is considered invalid if
   // a particle with (theta,phi) = (0,0) is outside the cut.

   // open min-max.dat file
   snprintf(filename, sizeof(filename), "%s/min-max.dat", dir);
   file.open((const char*) filename, ios::in);
   if (!file.is_open()) {
      Error("Dat2Root","Can't open %s: aborting",filename);
      return;
   }

   while (file.good()) {
      file >> ydum >> ddum >> tmin >> tmax >> phimin >> phimax;
      AddLine(ydum, ddum, -1,  0,  tmax  );
      AddLine(ydum, ddum,  1,  0, -tmin  );
      AddLine(ydum, ddum,  0, -1,  phimax);
      AddLine(ydum, ddum,  0,  1, -phimin);
      
      // Cut at (y,d) is invalid if (y,d,0,0) is outside of cut
      if (Eval(ydum, ddum, 0, 0) <= 0) {
         DiscardCut(ydum, ddum);
      }
   }
   file.close();
   file.clear();
}

//-------------------------------------------------------------------------

  /*
   * Example: TDatime date = ParseDate("Apr", 11, "06:31:01", 1975);
   */

TDatime THaRfn::ParseDate(const char* m, int d, const char* t, int y)
{
   int year=y, month=0, day=d, hour=0, min=0, sec=0;
   string monthstr = m;
   string timestr = t;
   string months[13] = { string("Bad"), string("Jan"), string("Feb"),
                         string("Mar"), string("Apr"), string("May"),
                         string("Jun"), string("Jul"), string("Aug"),
                         string("Sep"), string("Oct"), string("Nov"),
                         string("Dec") };

   // The day and year fields are already integers, so nothing needs
   // to be done for them.

   // Set month
   for (int i=1; i < 13; i++) {
      if (months[i] == monthstr)
         month = i;
   }

   // Set the hours
   unsigned int j=0, k=0;
   while (j < timestr.size()) {
      if (timestr[j] == ':')
         break;
      hour *= (unsigned int) TMath::Power(10,k);
      hour += (unsigned int) timestr[j] - 48;
      j++;
      k++;
   }
      
   // Set the minutes
   k=0;
   j++;
   while (j < timestr.size()) {
      if (timestr[j] == ':')
         break;
      min *= (unsigned int) TMath::Power(10,k);
      min += (unsigned int) timestr[j] - 48;
      j++;
      k++;
   }

   // Set the seconds
   k=0;
   j++;
   while (j < timestr.size()) {
      sec *= (unsigned int) TMath::Power(10,k);
      sec += (unsigned int) timestr[j] - 48;
      j++;
      k++;
   }
      
   return TDatime(year, month, day, hour, min, sec);
}

//-------------------------------------------------------------------------

   /*
    * Generate filename for rfn in the THaRfnModule expected format. This
    * format is used to save time when initilaizing the THaRfnModule, as it
    * takes a strangely long time to load the THaRfn objects into memory when
    * the module is looking for the correct one to use (from a file containing
    * multiple THaRfn objects). Thus the information about the date the rfn
    * lines were defined, collimator in/out, and spectrometer arm left/right
    * are encoded into the name of the THaRfn object, so that the module needs
    * only to look through the keys in the TFile to select the correct one.
    */  

TString THaRfn::MakeFileName() const
{
   char tmp[14];
   snprintf(tmp,sizeof(tmp),"%i",GetDate().GetDate()*100+GetDate().GetHour());
   
   TString name(tmp);
   name += (GetCollimator() == kIN) ? "-1" : "-0";
   name += (GetArm() == kRIGHT) ? "-1" : "-0";

   return name;
}
