/***************************************************************************

    filename             : THaRfn.h
    description          : 
    created              : Sun Jul 20 2003
    version              : 1.0
    author               : Micha D. Niskin
    email                : mniskin@yahoo.com

    Copyright (C) 2003 Micha D. Niskin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

 ***************************************************************************/


#ifndef FIU_THARFN
#define FIU_THARFN


#include <TNamed.h>
#include <TDatime.h>
#include "TRBTree.h"
#include "TRBNode.h"
#include <string>
#include "TRfn.h"


class THaRfn : public TNamed {

public:
   enum ECollimator_t { kOUT=0, kIN };
   enum EArm_t { kLEFT=0, kRIGHT };

   THaRfn();
   THaRfn(const char* name, const char* title, ECollimator_t collimator,
          EArm_t arm, Double_t dy, Double_t dd);

   void           AddLine(Double_t y, Double_t dp, Double_t a, Double_t b, 
                          Double_t c);
   void           DiscardCut(Double_t y, Double_t dp);
   Double_t       Eval(Double_t y, Double_t dp, Double_t th, Double_t ph);
   EArm_t         GetArm()        const { return fArm;                  }
   ECollimator_t  GetCollimator() const { return fCollimator;           }
   TDatime        GetDate()       const { return fDate;                 }
   Double_t       GetDd()         const { return fDd;                   }
   Double_t       GetDy()         const { return fDy;                   }
   // To obtain the partial derivatives at a point (y,d,th,ph) you must
   // first evaluate the RFN there. During the evaluation process these
   // derivatives and errors are calculated.
   Double_t       GetDR_DY()      const { return fDR_DY;                }
   Double_t       GetDR_DD()      const { return fDR_DD;                }
   Double_t       GetDR_DTh()     const { return fDR_DTh;               }
   Double_t       GetDR_DPh()     const { return fDR_DPh;               }
   Double_t       GetErr(Double_t dy, Double_t dd, Double_t dth, 
                         Double_t dph) const;
   Bool_t         IsValid()       const { return fIsValid;              }
   void           LoadDefs(const char* dir);
   TString        MakeFileName() const;
   void           SetDate(const TDatime& date) { fDate = TDatime(date); }
   void           SetIsValid(Bool_t state=kTRUE) { fIsValid = state;    }
   Double_t       operator()(Double_t y, Double_t dp, Double_t th,
                             Double_t ph) { return Eval(y, dp, th, ph); }

private:
   TDatime        ParseDate(const char* m, int d, const char* t, int y);

   TDatime        fDate;         // Date (as in information.dat)
   ECollimator_t  fCollimator;   // Collimator state: kIN, or kOUT
   EArm_t         fArm;          // Spectrometer arm: kLEFT, or kRIGHT
   Double_t       fDy;           // Grid spacing in y
   Double_t       fDd;           // Grid spacing in delta
   Double_t       fDR_DY;        // Partial derivative of Rfn w/resp. to y
   Double_t       fDR_DD;        // Partial derivative of Rfn w/resp. to delta
   Double_t       fDR_DTh;       // Partial derivative of Rfn w/resp. to theta
   Double_t       fDR_DPh;       // Partial derivative of Rfn w/resp. to phi
   TRBTree        fGrid;         // Binary tree containing TRfns for all y, dp
   Bool_t         fIsValid;      // Indicates whether this obj is ready to use

   ClassDef(THaRfn,1)
};

//-------------------------------------------------------------------------

inline Double_t THaRfn::GetErr(Double_t dy, Double_t dd, Double_t dth, 
                               Double_t dph) const
{
   Double_t drdy=GetDR_DY(), drdd=GetDR_DD(), drdth=GetDR_DTh(), 
            drdph=GetDR_DPh();

   return( TMath::Sqrt(drdy*dy*drdy*dy + drdd*dd*drdd*dd + drdth*dth*drdth*dth +
                      drdph*dph*drdph*dph) );
}
            
#endif
