/***************************************************************************

    filename             : THaRfnModule.C
    description          : 
    created              : Thu Aug 07 2003
    version              : 1.0
    author               : Micha D. Niskin
    email                : mniskin@yahoo.com

    Copyright (C) 2003 Micha D. Niskin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

 ***************************************************************************/


#include "THaRfnModule.h"
#include <THaVar.h>
#include <THaVarList.h>
#include <THaGlobals.h>
#include <TFile.h>
#include <TKey.h>
#include <TROOT.h>
#include <vector>
#include <string>

using std::vector;
using std::string;
using std::cout;
using std::endl;

ClassImp(THaRfnModule)


THaRfnModule::THaRfnModule(const char *name, const char *description) : 
            THaPhysicsModule(name, description)
{
   Info("THaRfnModule","Executing...");
   // Normal constructor.
   fRfn_R = 0;
   fRfn_L = 0;
}

//--------------------------------------------------------------------------

THaRfnModule::~THaRfnModule()
{
   Info("~THaRfnModule","Executing...");
   // Normal destructor.
}

//--------------------------------------------------------------------------

THaAnalysisObject::EStatus THaRfnModule::Init(const TDatime & run_time)
{
   Info("Init","Executing...");
   if (THaPhysicsModule::Init(run_time) != kOK)
       return fStatus;

   return fStatus;
}

//--------------------------------------------------------------------------

Int_t THaRfnModule::ReadDatabase(const TDatime& date)
{
   Info("ReadDatabase", "Executing...");
   
   string col_str, dir, rfn_file;
   vector<string> dirlist;
   THaRfn::ECollimator_t collimator;
   Int_t date_as_int = date.GetDate()*100 + date.GetHour();
   
   FILE *file = OpenRunDBFile(date);
   if( !file ) {
      Error("ReadDatabase", "Can't open db_run.dat file");
      return kFileError;
   }

   // extract collimator (IN/OUT) from the 'db_run.dat' file
   string tag(fPrefix); tag.append("collimator");
   LoadDBvalue( file, date, tag.c_str(), col_str );
   fclose(file);

   if (!col_str.size()) {
      Error("ReadDatabase", "Can't find 'Rfn.collimator' configuration");
      return kInitError;
   }
   
   collimator = ( col_str == "IN" ) ? THaRfn::kIN : THaRfn::kOUT;

   // figure out the path to the dir with the db files for this run
   dirlist = GetDBFileList("Rfn", date, "THaRfnModule::ReadDatabase()");
   if (!dirlist.size()) {
      Error("ReadDatabase", "Can't find DB files for %s", date.AsString());
      return kFileError;
   }

   // dirlist will contain the db file names. we just want the equivalent
   // of the unix 'dirname' command here, so we truncate the string at the
   // last '/' character
   dir = dirlist.back();
   dir = dir.erase(dir.find_last_of('/'), dir.size());

   // we expect to find the root file at '$DB_DIR/<date>/rfn.root'.
   rfn_file = dir + "/rfn.root";

   // open the root file
   TFile rootfile(rfn_file.c_str(), "READ");
   if (!rootfile.IsOpen()) {
      Error("ReadDatabase", "Can't open root file '%s'", rfn_file.c_str());
      return kFileError;
   }

   // find the THaRfn objects for left and right arms. The one with the
   // latest date <= 'date' above is chosen.
   string found_left(""), found_right("");
   TIter next(rootfile.GetListOfKeys());
   while (TObject *obj = next()) {
      TKey *key = (TKey*) obj;

      // parse the filename-- filename contains all info regarding date,
      // collimator, and which spectrometer (left or right arm)
      string name(key->GetName());
      Int_t date = atoi(name.substr(0,10).c_str());
      Int_t coll = atoi(name.substr(11,11).c_str());
      Int_t arm  = atoi(name.substr(13,13).c_str()); 

      // if the cut definitions are for the wrong collimator position
      // then we aren't interested in them
      if (coll != (Int_t) collimator)
         continue;

      // here we set a pointer to the rfn pointer corresponding to the
      // correct arm (e.g., if the cut defs are for the left arm, we set
      // the pointer to fRfn_L)
      THaRfn** rfn_to_set = (arm == (Int_t) THaRfn::kLEFT) ? &fRfn_L : &fRfn_R;
      string*  found_name = (arm == (Int_t) THaRfn::kLEFT) ? &found_left : &found_right;
      
      // default values for date diffs to ensure that the pointers do get set
      // even if there are no cut definitions from a time earlier than the
      // run date
      int date_diff_new = 0;
      int date_diff_old = 1;
      
      // if there's already a pointer set then we calculate diffs
      if (*rfn_to_set) {
         date_diff_new = date_as_int - date;
         date_diff_old = date_as_int - (((THaRfn*) *rfn_to_set)->GetDate().GetDate()*100 + ((THaRfn*) *rfn_to_set)->GetDate().GetHour());
      }

      // we're prejudiced toward older cut definitions, so...
      if (date_diff_old < 0 && date_diff_new > 0)
         continue;

      // here we set the pointer if we need to
      if (TMath::Abs(date_diff_new) < TMath::Abs(date_diff_old))
         *found_name = name.c_str();
   }         

   fRfn_L = (THaRfn*) rootfile.Get(found_left.c_str());
   fRfn_R = (THaRfn*) rootfile.Get(found_right.c_str());
      
   if (!fRfn_L)
      Error("ReadDatabase", "Didn't get RFN object for left arm");
   if (!fRfn_R)
      Error("ReadDatabase", "Didn't get RFN object for right arm");

   return 0;
}

//--------------------------------------------------------------------------

Int_t THaRfnModule::Process(const THaEvData & evdata)
{
   Double_t r_y  = ((THaVar*) gHaVars->Find("R.tr.tg_th"))->GetValue();
   Double_t r_d  = ((THaVar*) gHaVars->Find("R.tr.tg_th"))->GetValue();
   Double_t r_th = ((THaVar*) gHaVars->Find("R.tr.tg_th"))->GetValue();
   Double_t r_ph = ((THaVar*) gHaVars->Find("R.tr.tg_th"))->GetValue();
   
   Double_t l_y  = ((THaVar*) gHaVars->Find("L.tr.tg_th"))->GetValue();
   Double_t l_d  = ((THaVar*) gHaVars->Find("L.tr.tg_th"))->GetValue();
   Double_t l_th = ((THaVar*) gHaVars->Find("L.tr.tg_th"))->GetValue();
   Double_t l_ph = ((THaVar*) gHaVars->Find("L.tr.tg_th"))->GetValue();

   fRfn_Val_L = (fRfn_L) ? fRfn_L->Eval(l_y, l_d, l_th, l_ph) : TRfn::kBAD_RFN;
   fRfn_Val_R = (fRfn_R) ? fRfn_R->Eval(r_y, r_d, r_th, r_ph) : TRfn::kBAD_RFN;

   return 0;
}

//--------------------------------------------------------------------------

Int_t THaRfnModule::DefineVariables( EMode mode )
{
   Info("DefineVariables","Executing...");
   if( mode == kDefine && fIsSetup ) return kOK;
   fIsSetup = ( mode == kDefine );

   RVarDef vars[] = {
      { "L",      "R-Function value for left arm",       "fRfn_Val_L" },
      { "R",      "R-Function value for right arm",      "fRfn_Val_R" },
      { 0 }
   };
   return DefineVarsFromList( vars, mode );
}

