/***************************************************************************

    filename             : TRBTree.C
    description          : 
    created              : Sun Jul 20 2003
    version              : 1.0
    author               : Micha D. Niskin
    email                : mniskin@yahoo.com

    Copyright (C) 2003 Micha D. Niskin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

 ***************************************************************************/


#include "TRBTree.h"


ClassImp(TRBTree)

//-------------------------------------------------------------------------

TRBTree::TRBTree() : TObject()
{
   kIsBlack = kTRUE;
   fData    = 0;
   fParent  = 0;
   fLeft    = 0;
   fRight   = 0;
   kIsOwner = kFALSE;
}

//-------------------------------------------------------------------------

  /*
   * The leaf nodes will always be "sentinel" nodes. These are just empty,
   * black colored nodes. This simplifies dealing with the pointers, en-
   * suring that a node containing an object will always have valid fLeft
   * and fRight pointers.
   */

TRBTree::TRBTree(TObject* obj) : TObject()
{
   kIsBlack = kTRUE;
   fData    = obj;
   fParent  = 0;
   fLeft    = new TRBTree;  // empty black sentinel node
   fRight   = new TRBTree;  // another sentinel node
   fLeft->fParent = this;
   fRight->fParent = this;
   kIsOwner = kFALSE;
}

//-------------------------------------------------------------------------

TRBTree::~TRBTree()
{
   Clear();

   if (fParent)
      (IsLeftChild()) ? fParent->fLeft = 0 : fParent->fRight = 0;
}

//-------------------------------------------------------------------------

  /*
   * Add an object to the tree. If a node is added which is equal (according
   * to its TObject::Compare(TObject*) method) then the existing node will be
   * overwritten and a warning message displayed to the console.
   */

void TRBTree::Add(TObject* obj)
{
   if (!obj->IsSortable()) {
      Error("Add","Attempt to add a non-sortable object to tree");
   } else if (!fData) {
      fData = obj;
      
      fLeft = new TRBTree;
      fLeft->fParent = this;
      fLeft->SetOwner(IsOwner());
      
      fRight = new TRBTree;
      fRight->fParent = this;
      fRight->SetOwner(IsOwner());
      
      kIsBlack = kFALSE;
      Balance();
      ((TRBTree*) GetRoot())->kIsBlack = kTRUE;
   } else if (fData->IsA() != obj->IsA()) {
      Error("Add","Attempt to add a %s to a tree of %ss", 
            obj->IsA()->GetName(), fData->IsA()->GetName());
   } else if (!fData->Compare(obj)) {
      Warning("Add","Adding node which is equal to an existing node (overwriting)");

      if (IsOwner())
         delete fData;

      fData = obj;
   } else {
      (fData->Compare(obj) < 0) ? fLeft->Add(obj) : fRight->Add(obj);
   }
}

//-------------------------------------------------------------------------

  /*
   * Add all the objects in a a tree to this tree.
   */

void TRBTree::Add(TRBTree* tree)
{
   if (tree->IsOwner() != IsOwner()) {
      Error("Add","Attempt to add tree with different ownership");
      return;
   }

   if (!tree->fData)
      return;

   Add(tree->fLeft);
   Add(tree->fRight);
   Add(tree->fData);
}

//-------------------------------------------------------------------------

  /*
   * Recursively clear out all objects stored in tree, deleting them if
   * kIsOwner is set to kTRUE.
   */

void TRBTree::Clear(Option_t* option)
{
   if (fLeft && fLeft->IsOnHeap())
      delete fLeft;

   if (fRight && fRight->IsOnHeap())
      delete fRight;

   if (IsOwner() && fData && fData->IsOnHeap())
      delete fData;
   fData = 0;
}

//-------------------------------------------------------------------------

  /*
   * This method is needed in order to sort this object.
   */

Int_t TRBTree::Compare(const TObject* obj) const
{
   if (obj->IsA() != IsA())
      Error("Compare","Attempt to compare with a %s", obj->IsA()->GetName());

   if ((obj->IsA() != IsA()) || !fData || !((TRBTree*)obj)->fData)
      return 0;
   return (fData->Compare(((TRBTree*)obj)->fData));
}

//-------------------------------------------------------------------------

  /*
   * This is used from the interpreter to completely remove tree.
   */

void TRBTree::Delete(Option_t* option)
{
   if (IsOnHeap()) {
      gInterpreter->DeleteGlobal(this);
      delete this;
   }
}

//-------------------------------------------------------------------------

  /*
   * Return pointer to an object stored in the tree.
   */

TObject* TRBTree::FindObject(const TObject* obj) const
{
   if (!fData)
      return 0;
   else if (!fData->Compare(obj))
      return fData;
   else
      return (fData->Compare(obj)<0) ? fLeft->FindObject(obj) : fRight->FindObject(obj);
}

//-------------------------------------------------------------------------

  /*
   * Return pointer to the tree node which contains a given object.
   */

TRBTree* TRBTree::FindSubtreeContaining(TObject* obj)
{
   if (!fData)
      return 0;
   else if (!fData->Compare(obj))
      return this;
   else
      return (fData->Compare(obj)<0) ? fLeft->FindSubtreeContaining(obj) : fRight->FindSubtreeContaining(obj);
}

//-------------------------------------------------------------------------

  /*
   * Return depth of this node's subtree, i.e., the length of the longest
   * path from this node to a leaf node (not sentinel node).
   */

Int_t TRBTree::GetDepth() const
{
   if (!fData)
      return 0;
   else
      return (TMath::Max(fLeft->GetDepth(), fRight->GetDepth()) + 1);
}

//-------------------------------------------------------------------------

  /*
   * Return pointer to the root node.
   */

TRBTree* TRBTree::GetRoot()
{
   TRBTree* node = this;
   while (node->fParent) {
      node = node->fParent;
   }
   return node;
}

//-------------------------------------------------------------------------

  /*
   * Return number of objects stored in this node's subtree.
   */

Int_t TRBTree::GetSize() const
{
   if (!fData)
      return 0;
   else
      return (fLeft->GetSize() + fRight->GetSize() + 1);
}

//-------------------------------------------------------------------------
  
  /*
   * Required in order to sort objects of this class.
   */

Bool_t TRBTree::IsEqual(const TObject* obj) const
{
   if (obj->IsA() != IsA())
      Error("IsEqual","Attempt to compare with a %s", obj->IsA()->GetName());

   return (Compare(obj) == 0);
}

//-------------------------------------------------------------------------

  /*
   * Remove an object from the tree. The object is not deleted unless
   * kIsOwner is set to kTRUE by calling TRBTree::SetOwner().
   */

TObject* TRBTree::Remove(TObject* obj)
{
   Info("Remove","Not implemented yet");
   return 0;
}

//-------------------------------------------------------------------------

  /*
   * Recursively set ownership of this node's subtree. As with the
   * TCollection classes, the objects stored in the tree will not be
   * deleted unless kIsOwner is set to kTRUE.
   */

void TRBTree::SetOwner(Bool_t enable)
{
   if (fData) {
      fLeft->SetOwner(enable);
      fRight->SetOwner(enable);
   }
   kIsOwner = enable;
}

//-------------------------------------------------------------------------

  /*
   * Write the tree to a file. The kSingleKey option is default. If this
   * option is overridden then each node in the tree will be written to
   * a separate root file.
   */

Int_t TRBTree::Write(const char* name, Int_t option, Int_t bsize)
{
   Int_t ret1, ret2, ret3;
   if ((option & TObject::kSingleKey) || !fData) {
      return TObject::Write(name, option, bsize);
   } else {
      ret1=fLeft->Write(name, option, bsize);
      ret2=fRight->Write(name, option, bsize);
      ret3=TObject::Write(name, option, bsize);
      return (ret1 + ret2 + ret3); 
   }
}

//-------------------------------------------------------------------------

  /*
   *
   *
   *
   *             /                                /
   *            a                                b
   *           # \                             #   #
   *          b   4     a->RightRotate()     c       a
   *         # \       ----------------->   / \     / \
   *        c   3                          1   2   3   4
   *       / \
   *      1   2                Legend: node below # is red
   *                                   node below / is black
   *                                   node below \ is black
   *
   *
   *     When applying transformations it is required that a pointer to
   *     the root node will always point to the root node.
   *
   */

void TRBTree::RightRotate()
{
   TObject* tmpdata = fData;
   TRBTree* tmpright= fRight;

   TRBTree* b = fLeft;
   TRBTree* c = fLeft->fLeft;

   fData  = b->fData;
   fLeft  = c;
   fLeft->fParent = this;
   fRight = b;

   b->fData   = tmpdata;
   b->fLeft   = b->fRight;
   b->fRight  = tmpright;
   b->fRight->fParent = b;
}

//-------------------------------------------------------------------------

  /*
   *
   *
   *         \                                  \
   *          a                                  b
   *         / #                               #   #
   *        1   b       a->LeftRotate()      a       c
   *           / #     ---------------->    / \     / \
   *          2   c                        1   2   3   4
   *             / \
   *            3   4 
   *
   *
   */

void TRBTree::LeftRotate()
{
   TObject* tmpdata = fData;
   TRBTree* tmpleft = fLeft;

   TRBTree* b = fRight;
   TRBTree* c = fRight->fRight;

   fData  = b->fData;
   fLeft  = b;
   fRight = c;
   fRight->fParent = this;

   b->fData  = tmpdata;
   b->fRight = b->fLeft;
   b->fLeft  = tmpleft;
   b->fLeft->fParent = b;
}

//-------------------------------------------------------------------------

  /*
   *
   *
   *           /                                        /
   *          a                                        c
   *         # \                                     #   #
   *        b   4     a->DoubleRightRotate()       b       a
   *       / #       ----------------------->     / \     / \
   *      1   c                                  1   2   3   4
   *         / \
   *        2   3
   *
   *
   */

void TRBTree::DoubleRightRotate()
{
   TObject* tmpdata  = fData;
   TRBTree* tmpright = fRight;

   TRBTree* b = fLeft;
   TRBTree* c = fLeft->fRight;

   fData  = c->fData;
   fRight = c;
   c->fParent = this;

   b->fRight = c->fLeft;
   b->fRight->fParent = b;

   c->fData   = tmpdata;
   c->fLeft   = c->fRight;
   c->fRight  = tmpright;
   c->fRight->fParent = c;
}

//-------------------------------------------------------------------------

  /*
   *
   *
   *           \                                         \
   *            a                                         c
   *           / #                                      #   #
   *          1   b       a->DoubleLeftRotate()       a       b
   *             # \     ---------------------->     / \     / \
   *            c   4                               1   2   3   4
   *           / \
   *          2   3
   *
   *
   */

void TRBTree::DoubleLeftRotate()
{
   TObject* tmpdata = fData;
   TRBTree* tmpleft = fLeft;

   TRBTree* b = fRight;
   TRBTree* c = fRight->fLeft;

   fData = c->fData;
   fLeft = c;
   fLeft->fParent = this;

   b->fLeft = c->fRight;
   b->fLeft->fParent = b;

   c->fData   = tmpdata;
   c->fRight  = c->fLeft;
   c->fLeft   = tmpleft;
   c->fLeft->fParent = c;
}

//-------------------------------------------------------------------------

  /*
   *
   *
   *           \                                   #
   *            a                                   a
   *          #   #        a->ColorFlip()         /   \
   *        b       c     --------------->       b     c
   *       / #     / \                          / #   / \
   *      1   2   3   4                        1   2 3   4
   *
   *             
   *      Note: Red leaf can be any of 1, 2, 3, or 4, and 'a' can be
   *            either a right or a left node. 
   *
   */
    
void TRBTree::ColorFlip()
{
   kIsBlack = !kIsBlack;
   fLeft->kIsBlack = !fLeft->kIsBlack;
   fRight->kIsBlack = !fRight->kIsBlack;
}

//-------------------------------------------------------------------------

  /*
   * True if x is a left child. False for right children and the root node.
   */

Bool_t TRBTree::IsLeftChild()
{
   return (fParent && (this == fParent->fLeft));
}

//-------------------------------------------------------------------------

  /*
   * Check whether configuration is one of the five cases diagrammed above.
   * If so, then the appropriate transformation is applied, and the
   * configuration is then checked recursively at the root of the resulting 
   * subtree.
   */

void TRBTree::Balance()
{
   if (!fParent || fParent->kIsBlack)
      return;

   TRBTree* a = fParent->fParent;
   TRBTree* b = fParent;
   TRBTree* c = this;

   if (a && a->fLeft && a->fRight && 
       !(a->fLeft->kIsBlack || a->fRight->kIsBlack)) {
      a->ColorFlip();
   } else if (a && (c->IsLeftChild() == b->IsLeftChild())) {
      (c->IsLeftChild()) ? a->RightRotate() : a->LeftRotate();
   } else if (a && (c->IsLeftChild() != b->IsLeftChild())) {
      (c->IsLeftChild()) ? a->DoubleLeftRotate() : a->DoubleRightRotate();
   }
   a->Balance();
}
